"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BaseController = void 0;
const express_1 = require("express");
class BaseController {
    constructor() {
        this.router = (0, express_1.Router)();
    }
    /**
     * @return {void | Promise<void>}
     */
    initialize() {
        for (let route of this.onInitializeRoutes()) {
            route = this.unpackRoute(route);
            if (this.router[route.method] && typeof this.router[route.method] === 'function')
                this.router[route.method](route.path, this.callHandler.bind(this, route.method, route.path, route.handler));
        }
    }
    /**
     * @param {TRoute} route
     * @return {TRouteObject}
     * @private
     */
    unpackRoute(route) {
        if (Array.isArray(route)) {
            const [method, path, handler] = route;
            return {
                method, path, handler
            };
        }
        return route;
    }
    /**
     * @param {RouteHttpMethod} method
     * @param {string} path
     * @param {RouteHandler} handler
     * @param {e.Request} req
     * @param {e.Response} res
     * @param {e.NextFunction} next
     * @return {Promise<void>}
     * @private
     */
    async callHandler(method, path, handler, req, res, next) {
        try {
            const resultOrPromise = handler(req, res, next);
            if (resultOrPromise && resultOrPromise instanceof Promise) {
                const result = await resultOrPromise;
                return this.success.apply(this, [req, res, result]);
            }
            else {
                return this.success.apply(this, [req, res, resultOrPromise]);
            }
        }
        catch (e) {
            return this.error.apply(this, [req, res, e]);
        }
    }
    /**
     * @param {e.Request} req
     * @param {e.Response} res
     * @param data
     * @return {never}
     * @private
     */
    success(req, res, data) {
        return res.status(200)
            .set('Content-Type', 'application/json')
            .send(this.onTransformSuccessResponse(data));
    }
    /**
     * @param {e.Request} req
     * @param {e.Response} res
     * @param {Error} error
     * @return {never}
     * @private
     */
    error(req, res, error) {
        return res.status(200)
            .set('Content-Type', 'application/json')
            .send(this.onTransformErrorResponse(error));
    }
    /**
     * @param data
     * @return {any}
     * @protected
     */
    onTransformSuccessResponse(data) {
        return {
            status: 'ok',
            data: data
        };
    }
    /**
     * @param {Error} error
     * @return {any}
     * @protected
     */
    onTransformErrorResponse(error) {
        return {
            status: 'error',
            error: {
                name: error.constructor.name,
                message: error.message
            }
        };
    }
}
exports.BaseController = BaseController;
