import {NextFunction, Request, Response, Router} from 'express';
import {IController} from './IController';
import {TRoute, TRouteObject} from './TRoute';
import {RouteHttpMethod} from './RouteHttpMethod';
import {RouteHandler} from './RouteHandler';

export abstract class BaseController implements IController{

    public readonly router: Router = Router();

    /**
     * @return {void | Promise<void>}
     */
    public initialize(): void|Promise<void>{
        for(let route of this.onInitializeRoutes()){
            route = this.unpackRoute(route);
            if(this.router[route.method] && typeof this.router[route.method] === 'function')
                this.router[route.method](route.path, this.callHandler.bind(this, route.method, route.path, route.handler));
        }
    }

    /**
     * @param {TRoute} route
     * @return {TRouteObject}
     * @private
     */
    private unpackRoute(route: TRoute): TRouteObject{
        if(Array.isArray(route)){
            const [method, path, handler] = route;
            return {
                method, path, handler
            };
        }

        return route as TRouteObject;
    }

    /**
     * @param {RouteHttpMethod} method
     * @param {string} path
     * @param {RouteHandler} handler
     * @param {e.Request} req
     * @param {e.Response} res
     * @param {e.NextFunction} next
     * @return {Promise<void>}
     * @private
     */
    private async callHandler(method: RouteHttpMethod, path: string, handler: RouteHandler, req: Request, res: Response, next: NextFunction): Promise<never>{
        try{
            const resultOrPromise = handler(req, res, next);
            if(resultOrPromise && resultOrPromise instanceof Promise){
                const result = await resultOrPromise;
                return this.success.apply(this, [req, res, result]);
            }else{
                return this.success.apply(this, [req, res, resultOrPromise]);
            }
        }catch(e){
            return this.error.apply(this, [req, res, e]);
        }
    }

    /**
     * @param {e.Request} req
     * @param {e.Response} res
     * @param data
     * @return {never}
     * @private
     */
    private success(req: Request, res: Response, data: any): never{
        return res.status(200)
            .set('Content-Type', 'application/json')
            .send(this.onTransformSuccessResponse(data)) as never;
    }

    /**
     * @param {e.Request} req
     * @param {e.Response} res
     * @param {Error} error
     * @return {never}
     * @private
     */
    private error(req: Request, res: Response, error: Error): never{
        return res.status(200)
            .set('Content-Type', 'application/json')
            .send(this.onTransformErrorResponse(error)) as never;
    }

    /**
     * @return {TRoute[]}
     * @protected
     */
    protected abstract onInitializeRoutes(): TRoute[];

    /**
     * @param data
     * @return {any}
     * @protected
     */
    protected onTransformSuccessResponse(data: any): any{
        return {
            status: 'ok',
            data: data
        };
    }

    /**
     * @param {Error} error
     * @return {any}
     * @protected
     */
    protected onTransformErrorResponse(error: Error): any{
        return {
            status: 'error',
            error: {
                name: error.constructor.name,
                message: error.message
            }
        };
    }

}


